from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = 'UTMConfig',


class UTMConfig(AppConfig):
    name = 'wcd_utm'
    verbose_name = pgettext_lazy('wcd_utm', 'UTM')



import urllib.parse
import json
from django.http import HttpRequest, QueryDict

from .conf import settings
from . import resolver


__all__ = (
    'resolve_from_url',
    'resolve_from_dict',
    'resolve_from_request',
)


def resolve_from_url(url: str, **kwargs):
    url = urllib.parse.urlparse(url)
    query_dict = QueryDict(url.query)

    return resolver.resolve(query_dict.lists(), **kwargs)


def resolve_from_dict(values: dict, **kwargs):
    return resolver.resolve(
        [
            (key, (
                list(value)
                if isinstance(value, (list, tuple, set)) else
                [value]
            ))
            for key, value in values.items()
        ],
        **kwargs,
    )


def resolve_from_request(request: HttpRequest, **kwargs):
    if settings.HEADER_ORIGIN_URL in request.META:
        return resolve_from_url(
            request.META[settings.HEADER_ORIGIN_URL], **kwargs
        )

    if settings.HEADER_JSON in request.META:
        return resolve_from_dict(
            json.loads(request.META[settings.HEADER_JSON]), **kwargs
        )

    return resolver.resolve(request.GET.lists(), *kwargs)



import logging

from .shortcuts import resolve_from_request
from . import session_storage


logger = logging.getLogger(__name__)


class UTMSessionMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        try:
            session_storage.put(request.session, resolve_from_request(request))
        except Exception as e:
            logging.exception(e, exc_info=True)

        response = self.get_response(request)
        return response



import pytest
from django.test import RequestFactory
from django.contrib.sessions.middleware import SessionMiddleware

from wcd_utm.middleware import UTMSessionMiddleware
from wcd_utm.conf import settings


@pytest.mark.django_db
def test_multiple_same_requests():
    url = '/?utm_content=5&utm_content=shaza&gclid=5'
    factory = RequestFactory()
    request = factory.get(url)
    md = UTMSessionMiddleware(lambda r: None)
    smd = SessionMiddleware(lambda r: None)

    smd.process_request(request)
    md(request)
    md(request)

    assert len(request.session[settings.SESSION_STORAGE_KEY]) == 1

    request1 = factory.get(url)
    request2 = factory.get('/path/to/something' + url + '&with=get&parameters')
    smd.process_request(request1)
    md(request1)
    request2.session = request1.session
    md(request2)

    assert len(request1.session[settings.SESSION_STORAGE_KEY]) == 1


@pytest.mark.django_db
def test_multiple_different_requests():
    url = '/?utm_content=5&utm_content=shaza&gclid=5'
    factory = RequestFactory()
    md = UTMSessionMiddleware(lambda r: None)
    smd = SessionMiddleware(lambda r: None)

    request1 = factory.get(url)
    request2 = factory.get(url + '&utm_content=third')
    smd.process_request(request1)
    md(request1)
    request2.session = request1.session
    md(request2)

    assert len(request1.session[settings.SESSION_STORAGE_KEY]) == 2



from typing import *
from dataclasses import dataclass, field

from px_settings.contrib.django import settings as setting_wrap


__all__ = 'Settings', 'settings',


@setting_wrap('WCD_UTM')
@dataclass
class Settings:
    RESOLVABLE_PREFIXES: Sequence[str] = field(default_factory=lambda: ['utm_'])
    PARAMETERS: Sequence[str] = field(default_factory=lambda: [
        'utm_source',
        'utm_medium',
        'utm_campaign',
        'utm_term',
        'utm_content',

        'gclid',  # Google ad click
        'aclk',  # Bing ad click
        'msclkid',  # MSFT ad click (non-Bing)
        'fbclid',  # Facebook ad click
        'twclid',  # Twitter ad click
    ])
    UNWRAP_PREFIXED_PARAMETERS: bool = False

    SESSION_STORAGE_KEY: str = 'utm_params_stored'
    SESSION_ACCESS_KEY: str = 'utm_params'

    HEADER_ORIGIN_URL: str = 'HTTP_X_UTM_ORIGIN_URL'
    HEADER_JSON: str = 'HTTP_X_UTM_JSON'


settings = Settings()



from typing import *
from collections import OrderedDict

from .conf import settings


__all__ = 'QueryLists', 'resolve'

QueryLists = List[Tuple[str, List[str]]]
Resolved = Dict[str, List[str]]


def _merge_lists(current: Optional[List[str]], merge: List[str]):
    current = [] if current is None else current
    return current + [v for v in merge if v not in current]


def resolve(
    lists: QueryLists,
    parameters: Sequence[str] = settings.PARAMETERS,
    prefixes: Sequence[str] = settings.RESOLVABLE_PREFIXES,
) -> Resolved:
    match_prefixes = [(k, len(k)) for k in prefixes]
    parameters = tuple(parameters)
    merged = OrderedDict()

    for key, values in lists:
        prefix, l = next(
            (x for x in match_prefixes if key.startswith(x[0])),
            (None, 0)
        )

        if key in parameters or prefix is not None:
            merged[key] = _merge_lists(merged.get(key, None), values)

        if prefix is not None and settings.UNWRAP_PREFIXED_PARAMETERS:
            subkey = key[l:]

            if subkey not in parameters:
                merged[subkey] = _merge_lists(merged.get(subkey, None), values)

    return merged



from typing import *
from django.contrib.sessions.backends.base import SessionBase
from django.utils import timezone

from .resolver import Resolved

from .conf import settings


__all__ = 'put', 'clear',

Stored = List[Tuple[str, Resolved]]


def put(session: SessionBase, params: Resolved) -> bool:
    if not params:
        return False

    storage_key = settings.SESSION_STORAGE_KEY
    access_key = settings.SESSION_ACCESS_KEY
    session.setdefault(storage_key, [])
    session.setdefault(access_key, {})

    storage = session[storage_key]

    if storage:
        _, latest = storage[-1]

        if latest == params:
            return False

    timestamp = timezone.now().isoformat()
    storage.append((timestamp, params))
    session[access_key] = params

    session.modified = True

    return True


def clear(session: SessionBase) -> Stored:
    session.pop(settings.SESSION_ACCESS_KEY)
    return session.pop(settings.SESSION_STORAGE_KEY, [])



import pytest
from django.test import RequestFactory

from wcd_utm import resolver
from wcd_utm.shortcuts import resolve_from_request, resolve_from_url
from wcd_utm.conf import settings


@pytest.mark.django_db
def test_default_parameters_resolver():
    url = '/?utm_content=5&utm_content=shaza&gclid=5'
    request = RequestFactory().get(url)
    resolved = resolver.resolve(request.GET.lists())

    assert resolved == resolve_from_request(request)
    assert resolved == resolve_from_url(url)

    assert resolved['gclid'] == ['5']
    assert resolved['utm_content'] == ['5', 'shaza']
    assert 'content' not in resolved
    assert len(resolved) == 2


@pytest.mark.django_db
def test_url_header():
    url = '/?utm_content=5&utm_content=shaza&gclid=5'
    request = RequestFactory().get(
        url, **{settings.HEADER_ORIGIN_URL: url + '&gclid=6'}
    )
    resolved = resolver.resolve(request.GET.lists())
    request_resolved = resolve_from_request(request)

    assert resolved != request_resolved
    assert request_resolved['gclid'] == ['5', '6']

    url = '/?utm_content=5&utm_content=shaza&gclid=5'
    request = RequestFactory().get(
        url, **{settings.HEADER_ORIGIN_URL: '/?gclid=6'}
    )
    resolved = resolver.resolve(request.GET.lists())
    request_resolved = resolve_from_request(request)

    assert resolved != request_resolved
    assert request_resolved['gclid'] == ['6']
    assert len(request_resolved) == 1


@pytest.mark.django_db
def test_json_header():
    url = '/?utm_content=5&utm_content=shaza&gclid=5'
    request = RequestFactory().get(
        url, **{settings.HEADER_JSON: '{"gclid": 8, "alambalam": "7"}'}
    )
    resolved = resolver.resolve(request.GET.lists())
    request_resolved = resolve_from_request(request)

    assert resolved != request_resolved
    assert request_resolved['gclid'] == [8]
    assert len(request_resolved) == 1
